__version__ = '2022.2.2'

from .camera.camera import Camera
from .camera.camera_calibration import CameraCalibration
from .camera.camera_configuration import CameraConfiguration
from .camera.camera_properties import CameraProperties

from .networktables_connection.network_tables_connection import NetworkTablesConnection

from .detectors.detector import Detector
from .detectors.haar_cascade_detector import HaarCascadeDetector
from .detectors.threshold_detector import ThresholdDetector

from .direction_modifiers.direction_modifier import DirectionModifier
from .direction_modifiers.stop_if_close_modifier import StopIfCloseModifier
from .directions.directing_functions import *
from .directions.director import Director

from .exceptions.exceptions import *
from .image_filters import kernels
from .image_filters.image_filter import IMAGE_FILTERS, image_filter
from .image_filters.image_filters import *

from .image_utilities.display_contours import display_contours
from .image_utilities.display_image import stitch_images, display_image
from .image_utilities.naming_conventions import *
from .image_utilities.open_image import open_image
from .image_utilities.photo_array import photo_array

from .ovl_math import *
from .morphological_functions import morphological_functions

from .target_filters.contour_filters import *
from .target_filters.predicate_target_filter import predicate_target_filter
from .target_filters.shape_filters.circle_filter import circle_filter
from .target_filters.shape_filters.constraining_circle_filter import constraining_circle_filter
from .target_filters.shape_filters.horizontal_rectangle_filter import horizontal_rectangle_filter
from .target_filters.shape_filters.polygon_filter import polygon_filter
from .target_filters.shape_filters.rotated_square_filter import rotated_square_filter
from .target_filters.shape_filters.straight_square_filter import straight_square_filter
from .target_filters.shape_filters.vertical_rectangle_filter import vertical_rectangle_filter
from .target_filters.shape_filters.rotated_rectangle_filter import rotated_rectangle_filter
from .target_filters.shape_filters.straight_rectangle_filter import straight_rectangle_filter
from .target_filters.shape_filters.triangle_filter import triangle_filter
from .target_filters.sorters import *
from .target_filters.target_filter import target_filter, TARGET_FILTERS

from .thresholds import *
from .thresholds.binary_threshold import BinaryThreshold
from .thresholds.binary_threshold import BinaryThresholdType
from .thresholds.canny_edge import CannyEdge
from .thresholds.color.built_in_colors import HSV
from .thresholds.color.color import Color
from .thresholds.color.multi_color import MultiColor

from .utils.constants import *
from .utils.team_number_to_ip import team_number_to_ip

from .visions.ambient_vision import AmbientVision
from .visions.multi_vision import MultiVision
from .visions.vision import Vision



import math
import types
from functools import reduce
from logging import getLogger
from typing import List, Union, Any, Callable, Tuple, Iterable

import cv2
import numpy as np

from ovl import OMIT_DIMENSION_VALUES, DEFAULT_FAILED_DETECTION_VALUE, VISION_LOGGER
from ..camera.camera import Camera, configure_camera
from ..camera.camera_configuration import CameraConfiguration
from ..detectors.detector import Detector
from ..directions.directing_functions import center_directions
from ..directions.director import Director
from ..exceptions.exceptions import CameraError, ImageError
from ..partials.filter_applier import apply
from ..thresholds.threshold import Threshold
from ..utils.constants import DEFAULT_IMAGE_HEIGHT, DEFAULT_IMAGE_WIDTH, BASE_LOGGER
from ..utils.get_function_name import get_function_name
from ..utils.types import Target
from ..utils.vision_detector_arguments import arguments_to_detector

logger = getLogger(f"{BASE_LOGGER}.{VISION_LOGGER}")


class Vision:
    """
    Vision object represents a computer vision pipeline.
    The pipeline consists of 4 main stages:
         1. processing - 'apply_all_image_filters' which uses a list of image_filter functions
         2. detection - 'detect' which comes from Detector objects and detects objects (targets, bounding rectangles..)
         3. filtering  - 'apply_target_filters' which is uses filter_functions like contour_filters
         4. conversion & usage - 'direct' which comes from director objects

    Each functionality can be used to easily create a complex yet modular

    Additional capabilities and tuning options are:
        Image filters (Blurs, rotations, cropping),
        Morphological functions,

        1. camera handling
        2. connection clean-up and sending

    Vision can also be used as a part of a more complex pipeline.

    MultiVision can contain multiple vision objects and switch between pipelines, allowing for very versatile logic
    that can fit multiple needs.

    Ambient Vision is another option for using 2 different Vision objects and alternate between the 2.
    """

    def __init__(self, detector: Detector = None, threshold: Threshold = None,
                 morphological_functions: List[types.FunctionType] = None,
                 target_filters: List[types.FunctionType] = None, director: Director = None,
                 target_selector: Union[int, tuple, Callable] = 1,
                 width=DEFAULT_IMAGE_WIDTH, height=DEFAULT_IMAGE_HEIGHT,
                 camera: Union[int, str, Camera, cv2.VideoCapture, Any] = None,
                 camera_configuration: CameraConfiguration = None, image_filters: List[types.FunctionType] = None,
                 ovl_camera: bool = False, haar_classifier: str = None, logger_name: str = None):
        """
        :param detector: a Detector object responsible for detecting targets
        :param threshold: threshold is a shortcut for detecting
        :param target_filters: the list of target_filter functions that remove targets that aren't what you want,
        can also perform grouping, sorting or any other action you want to
        :param director: a functions that receive a list or a single contour and returns director
        :param width: the width (in pixels) of images taken with the camera
        :param height: the height (in pixels)
        :param camera: a Camera object (cv2.VideoCapture, ovl.Camera) or source from which to open a camera
        :param camera_configuration: Special camera settings like calibration or offset used for
                                image correction and various direction calculations.
        :param image_filters: a list of image altering functions that are applied on the image.
        :param ovl_camera: a boolean that makes the camera opened to be ovl.Camera instead of cv2.VideoCapture
        :param haar_classifier:
        :param target_selector: decides how many/what targets are selected after targets have been filtered
        """
        if not (detector is None and threshold is None and haar_classifier is None):
            mutually_exclusive_arguments = {"threshold": (threshold, morphological_functions),
                                            "detector": (detector,),
                                            "haar_cascade": (haar_classifier,)}

            detector = arguments_to_detector(mutually_exclusive_arguments)
        self.detector = detector

        self.width = width
        self.height = height
        self.target_filters = target_filters or []
        self.director = director or Director(center_directions,
                                             failed_detection=DEFAULT_FAILED_DETECTION_VALUE,
                                             target_selector=target_selector)
        self.image_filters = image_filters or []
        self.camera = None
        self.camera_port = None
        self.camera_configuration = camera_configuration
        self.logger = getLogger(logger_name or VISION_LOGGER)

        if isinstance(camera, (cv2.VideoCapture, Camera)) or camera is None:
            self.camera = camera
        else:
            self.camera_setup(camera, width, height, camera_configuration, ovl_camera=ovl_camera)

    def __repr__(self):
        return str(self)

    def __str__(self):
        filters = [get_function_name(filter_function) for filter_function in self.target_filters]
        image_filters = [get_function_name(image_filter) for image_filter in self.image_filters]
        return f"Vision: \n Detector: {self.detector} \n Filters: {filters} \n Image Filters: {image_filters}"

    @property
    def target_selector(self):
        """
        The wanted amount of targets
        Determined by `self.director`
        (0 None or math.inf if there is no limit, 1 if 1 target is wanted etc.)

        """
        if self.director is None:
            return math.inf
        return self.director.target_selector

    def get_image(self) -> np.ndarray:
        """
        Gets an image from `self.camera` and applies image filters

        :return: the image
        :raises: ImageError if the image fau

        """
        if self.camera is None:
            raise CameraError("No camera given, (Camera is None)")
        if not self.camera.isOpened():
            raise CameraError("The Vision's camera is not open (Has it been closed or disconnected?)")
        output = self.camera.read()
        if len(output) == 2:
            success, image = output
            if not success:
                raise ImageError("Failed to take image")
            return image
        else:
            return output

    def apply_target_filter(self, filter_function, targets):
        """
        Applies a filter function on the target list, this is used to remove targets
        that do not match desired features

        NOTE: `Vision.detect` is mainly used for full object detection and filtering,
        refer to it for common use of Vision

        :param filter_function: Filter functions are functions that take out targets that
         for example: straight_rectangle_filter removes contours that are not rectangles that are parallel
         to the frame of the picture
        :param targets: the targets on which the filter should be applied (list of numpy.ndarrays or bounding boxes,
        depends on the values returned by your detector)
        :return: returns the output of the filter function.

        """
        name = get_function_name(filter_function)
        self.logger.info(f'Before "{name}": {len(targets)}')
        filtered_targets = filter_function(targets)
        return filtered_targets

    def apply_target_filters(self, targets: Iterable["Target"]) -> Iterable["Target"]:
        """
        Applies all target filters on a list of targets, one after the other.
        Applies the first filter and passes the output to the second filter,

        :param targets: List of targets (numpy arrays or bounding boxes) to
        :return: a list of all ratios given by the filter functions in order.

        """
        return list(reduce(apply, self.target_filters, targets))

    def apply_image_filters(self, image: np.ndarray) -> np.ndarray:
        """
        Applies all given image filters to the given image
        This is used to apply various image filters on your image in a pipeline,
        like blurs, image cropping, contrasting, sharpening, rotations, translations etc.

        :param image: the image that the image filters should be applied on (numpy array)
        :return: the image with the filters applied
        """
        return reduce(apply, self.image_filters, image)

    def get_directions(self, targets: Iterable["Target"], image: np.ndarray) -> Any:
        """
        Calculates the directions, based on targets found in the given image

        :param targets: final targets after filtering
        :param image: the image
        :return: returns the direction
        """
        return self.director.direct(targets, image)

    def camera_setup(self, source=0, image_width=None, image_height=None,
                     camera_configuration: CameraConfiguration = None, ovl_camera=False):
        """
        Opens up the camera reference and sets a given width and height to all images taken

        :param image_width: the width of the images to be taken, 0 does not set a width
        :param image_height: the height of the images to be taken, 0 does not set a height
        :param camera_configuration: a camera configuration object that allows easy camera configuration of the various
        properties defined in `cv2.VideoCapture`. Depends on the camera you use and its driver.
        :param source: the location from which to open the camera
         string for network connections int for local USB connections.
        :param ovl_camera: if the camera object should be ovl.Camera
        :return: the camera object, also sets `self.camera`.
        """

        image_height = image_height or self.height
        image_width = image_width or self.width
        self.camera_port = source
        if ovl_camera:
            camera = Camera(source=source, image_width=image_width, image_height=image_height)
            self.width = image_width
            self.height = image_height

        else:
            camera = cv2.VideoCapture(source)
            if image_width not in OMIT_DIMENSION_VALUES:
                camera.set(cv2.CAP_PROP_FRAME_WIDTH, image_width)
                self.width = image_width
            if image_height not in OMIT_DIMENSION_VALUES:
                camera.set(cv2.CAP_PROP_FRAME_HEIGHT, image_height)
                self.height = image_height

        if not camera.isOpened():
            raise CameraError(f"Camera did not open correctly! Make sure it's connected, source: {self.camera_port}")
        if camera_configuration is not None:
            configure_camera(camera, camera_configuration)
        self.camera = camera
        return camera

    def detect(self, image, *args, **kwargs) -> Tuple[Iterable["Target"], "np.ndarray"]:
        """
        This is the function that performs processing, detection and filtering on a given image, essentially passing
        the image through the detection related part of the pipeline

        `detect` applies image filters, detects objects in the filtered images
        (using the passed/created detector object)
        and finally applies all the target_filters on the image.

        args and kwargs are passed to the detect function (passed to the detect method of the detector)

        :param image: image in which the vision should detect an object
        :return: targets and the filtered image

        """
        filtered_image = self.apply_image_filters(image)
        targets = self.detector.detect(filtered_image, *args, **kwargs)
        filtered_targets = self.apply_target_filters(targets)
        return filtered_targets, filtered_image



from enum import Enum

import cv2


class CameraProperties(int, Enum):
    """
    An enum containing the various properties that can be configured for cameras,
    Different cameras support different apis, this means not all properties are supported for each camera, checking with
    the camera manufacturer and the drivers written in opencv determines what is supported,
    you're encouraged to test what options work for your camera beforehand.

    You can read more about each option at the opencv documentation here:

    Examples:


    Simply use the enum members as the keys to set various properties:

    e.g. setting camera exposure:

    .. code-block:: python

        import ovl

        camera = ovl.Camera(0)
        camera.set(CameraProperties.AUTO_EXPOSURE, 0)
        camera.set(CameraProperties.EXPOSURE, -12)


    Camera properties work both with ovl's `Camera` and with opencv's `VideoCapture`

    .. code-block:: python

        import cv2
        from ovl import CameraProperties

        camera = cv2.VideoCapture(0)

        camera.set(CameraProperties.IMAGE_WIDTH, 320)
        camera.set(CameraProperties.IMAGE_HEIGHT, 240)



    """
    IMAGE_WIDTH = cv2.CAP_PROP_FRAME_WIDTH
    IMAGE_HEIGHT = cv2.CAP_PROP_FRAME_HEIGHT
    CAMERA_FPS = cv2.CAP_PROP_FPS
    CAMERA_MODE = cv2.CAP_PROP_MODE
    BRIGHTNESS = cv2.CAP_PROP_BRIGHTNESS
    CONTRAST = cv2.CAP_PROP_CONTRAST
    SATURATION = cv2.CAP_PROP_SATURATION
    HUE = cv2.CAP_PROP_HUE
    GAIN = cv2.CAP_PROP_GAIN
    EXPOSURE = cv2.CAP_PROP_EXPOSURE
    MONOCHROME = cv2.CAP_PROP_MONOCHROME
    SHARPNESS = cv2.CAP_PROP_SHARPNESS
    AUTO_EXPOSURE = cv2.CAP_PROP_AUTO_EXPOSURE



from pathlib import Path

import cv2


def open_image(image_path, flags=None):
    """
    Opens an image file, has a more direct error detection to find out problems
    that occurred during image opening, has a cost in performance.
    
    :param image_path: the path of the image to open
    :param flags: there are multiple flags that can be used to alter how the image is loaded
     https://docs.opencv.org/4.3.0/d4/da8/group__imgcodecs.html#ga288b8b3da0892bd651fce07b3bbd3a56
    :return: the opened image or an iterator that opens the list of images given
    """
    if not Path(image_path).is_file():
        raise OSError(f"Given image at path {image_path} does not exist or is not a file!")

    image = cv2.imread(image_path, flags=flags)
    if image is None:
        raise ValueError(f"Failed to read the image file, Check that {image} exists, that it's a valid format and that "
                         f"you have the proper permissions to read it.")
    return image



DEFAULT_IMAGE_HEIGHT = 240
DEFAULT_IMAGE_WIDTH = 320
DEFAULT_CAMERA_SOURCE = 0
MIN_OPENCV_CAMERA_PROPERTY = 0
MAX_OPENCV_CAMERA_PROPERTY = 18
DEFAULT_KERNEL_SIZE = (3, 3)
OMIT_DIMENSION_VALUES = (-1, 0)
DEFAULT_FAILED_DETECTION_VALUE = "Failed to detect targets"
VISION_LOGGER = "vision"
MULTI_VISION_DELAY = 0.1
BASE_LOGGER = "ovl"
MULTIVISION_LOGGER = "multivision"



import cv2

from ..predicate_target_filter import predicate_target_filter
from ...ovl_math.shape_fill_ratios import triangle_fill_ratio
from ...utils.types import RangedNumber


@predicate_target_filter
def triangle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.8,
                    approximation_coefficient: RangedNumber(0, 1) = 0.02):
    """
    Receives a list of contours and returns only those that are approximately
    triangle and have 3 sides.

    :param approximation_coefficient: the approximation coefficient affects
    how the contour's perimeter is approximated, the larger the number is (out of 1) the more the approximation of
    the contour is a line
    :param contour: contour to be evaluated
    :param min_area_ratio: the minimum ratio between the area of the contour and the bounding triangle
    :return: the filtered list
    """
  
    fill_ratio = triangle_fill_ratio(contour)
    peri = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, approximation_coefficient * peri, True)
    return fill_ratio > min_area_ratio and len(approximation) == 3
  


import cv2
import numpy as np

from .detector import Detector


class HaarCascadeDetector(Detector):
    """
    A detector used to detect objects using haar cascade algorithm
    The Detector initializes using a xml file containing the descriptor

    The Detector uses the underlying cv2.CascadeClassifier

    For more information on

    """
    def __init__(self, classifier: str):
        self.classifier_source = classifier
        self.classifier = cv2.CascadeClassifier(classifier)

    def detect(self, image: np.ndarray, *args, **kwargs):
        greyscale = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        return self.classifier.detectMultiScale(greyscale)



from ...ovl_math.shape_fill_ratios import circle_fill_ratio
from ..predicate_target_filter import predicate_target_filter
from ...utils.types import RangedNumber


@predicate_target_filter
def circle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.7):
    """
    Filters out contour which are not approximately circle.
    
    :param contour: list of contours (numpy arrays) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
                         circle and the contour (contour /enclosing circle)
    :return: the list of contours that fit the criteria
    """

    fill_ratio, _ = circle_fill_ratio(contour)
    return fill_ratio >= min_area_ratio



import cv2

from ..predicate_target_filter import predicate_target_filter
from ...ovl_math.shape_fill_ratios import rectangle_fill_ratio_straight
from ...utils.types import RangedNumber


@predicate_target_filter
def horizontal_rectangle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.8):
    """
    Receives a contour and return True only those that are approximately a horizontal rectangle

    :param contour: contour (numpy.ndarray) to be filtered
    :param min_area_ratio
    :return: the contour list filtered.
    """
    fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(contour)
    perimeter = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, 0.02 * perimeter, True)
    return fill_ratio > min_area_ratio and contour_width > contour_height and len(approximation) == 4



from typing import *

import cv2
import numpy as np

from ..threshold import Threshold

BaseForColor = NewType("BaseForColor", Union[int, Tuple[Tuple[int, int, int], Tuple[int, int, int]]])
SERIALIZED_COLOR_KEYS = {"high", "low"}


def validate(value, ceiling):
    """
    Checks if val is positive and less than the ceiling value.

    :param value: The value to be checked (usually an int)
    :param ceiling: The highest value "val" can be. (usually an int)
    :return: True if val is less than ceiling and not negative
    """
    value = int(value)
    return 0 <= value < ceiling


def assert_hsv(hsv_point):
    """
    Makes sure the hsv point(or vector) given in the parameter has valid values.

    :param hsv_point: a 3 length list with 3 ints describing a point the hsv color space
    that describe a color limit in a range for a findContour function.
    :return: True if valid False otherwise
    :rtype: bool
    """

    return validate(hsv_point[0], 180) and validate(hsv_point[1], 256) and validate(hsv_point[2], 256)


def _range_assemble(color_range, padding):
    if isinstance(color_range, int):
        return [color_range, *padding]
    if isinstance(color_range, tuple):
        return list(color_range)
    return color_range


class Color(Threshold):
    """
    Color is a Threshold object (an object that turns an image to a binary image
    - an image with pixels with the value 1 and 0)
    Color object thresholds using 2 HSV color ranges.

    Read more about HSV here:
    https://www.lifewire.com/what-is-hsv-in-design-1078068

    HSV Ranges in OVL (And in the underlying open-cv (cv2)) are:
    Hue - 0 to 179 (360/2 so divide any regular HSV value you use by 2)

    So if Yellow is 40 - 80 (degrees) in regular HSV palette its 20 to 40 in ovl
    Saturation - 0 to 255 (255 is equal to 1 or 100%)
    Value - 0 to 255 (255 is equal to 1 or 100%)

    2 ranges are passed to the Constructor:
        1) Low HSV Range - the lowest acceptable H S V values ([low_H, low_S, low_V])
        2) High HSV Range - the highest acceptable H S V values ([high_H, high_S, high_V])

    .. code-block:: python

        low_range = [15, 100, 100]
        high_range = [45, 255, 255]
        color = ovl.Color(low_range, high_range)



    `Color` can be passed to a `Vision` to threshold binary images
    Threshold object can be used by themselves using the `color.convert()` method.

    NOTE: Threshold objects automatically convert images to HSV (From the default BGR)

    There are multiple built-in "battery included" pre-made color object
    for instant use in testing and tuning look at the `HSV` or more information
    """

    def validate(self, *args, **kwargs):
        return assert_hsv(self.low_bound) and assert_hsv(self.high_bound)

    def __init__(self, low: BaseForColor, high: BaseForColor):
        """
         Constructor for the Color used to turn images to binary images based on
         HSV color space (pixels that are in the specified HSV Range)
         H - 0 - 179
         S - 0 - 255
         V - 0 - 255

        :param high: high hsv limit of the color
        :param low: low hsv limit of the color
        """

        low = _range_assemble(low, [100, 100])
        high = _range_assemble(high, [255, 255])
        self.__low_bound = np.array(low)
        self.__high_bound = np.array(high)

    def threshold(self, image: np.ndarray) -> np.ndarray:
        return cv2.inRange(image, self.low, self.high)

    def convert(self, image: np.ndarray) -> np.ndarray:
        """
        Converts a given image to hsv and then thresholds and returns the binary mask

        :param image: a BGR image
        :return: binary mask
        :rtype: numpy array
        """
        hsv_image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        return self.threshold(hsv_image)

    def __repr__(self):
        return 'Color({}, {})'.format(repr(self.low_bound), repr(self.high_bound))

    @property
    def low_bound(self):
        if isinstance(self.__low_bound, np.ndarray):
            return self.__low_bound.tolist()
        return self.__low_bound

    @property
    def high_bound(self):
        if isinstance(self.__high_bound, np.ndarray):
            return self.__high_bound.tolist()
        return self.__high_bound

    @property
    def low(self):
        """
        Returns the low hsv limit of the color.

        :return: An uint8 numpy array with the low limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__low_bound

    @property
    def high(self):
        """
        Returns the high hsv limit of the color.

        :return: An uint8 numpy array with the high limit of the color.
        :rtype: uint8 numpy array
        """
        return self.__high_bound

    def __str__(self):
        return self.__repr__()



import functools


def get_function_name(func):
    """
    Get the name of a function.
    """
    try:
        if isinstance(func, functools.partial):
            return func.func.__name__
        return func.__name__
    except AttributeError:
        return repr(func)



import math
from logging import getLogger
from typing import *

import numpy as np

from ..exceptions.exceptions import NotEnoughTargets
from .target_selector import validate_target_selector
from ..direction_modifiers.direction_modifier import DirectionModifier

DEFAULT_DIRECTOR_LOGGER = "Director"


class Director:
    """
    Directors are object that extract usable information from the objects that were detected image.
    This information be where they are located in the image relative to the center (directions for a robot),
    how many objects were detected, identifying an object or any other logical usage of the object detection
    rather than raw image processing or computer vision.
    This is the final step before sending/using the data.

    Director.direct works in the following way:
    - if the amount of targets is less than target amount then set directions to failed_detection
    - else it passes only target_selector of targets to the direction function
    - The final result (direction function or failed_detection) is passed to the direction monitors
    - each modify_directions is applied to the result and its result is passed to the next one
    - then the final result is returned
    """

    def __init__(self, directing_function: Callable, failed_detection: Any = None, target_selector: int = 0,
                 direction_modifiers: List[DirectionModifier] = None, logger: str = None):
        """
        :param directing_function: the function that performs the initial direction calculation
        :param failed_detection: The value returned on a failed detection
        :param target_selector: selects targets after the targets have been filtered by the target filters,
        can be a number, range or selecting function
        (selecting functions receive the targets and should return False for if it failed to select)
        :param direction_modifiers: list of DirectionMonitor objects that alter the directions to be sent
        """
        self.directions = directing_function
        self.target_selector = validate_target_selector(target_selector)
        self.direction_monitors = direction_modifiers
        self.failed_detection = failed_detection
        self.logger = getLogger(logger or DEFAULT_DIRECTOR_LOGGER)

    def _inf_select(self, targets: List, selector):
        if math.isinf(selector):
            selector = None
        return targets[:selector]

    def _limit_selector(self, targets: List):
        if len(targets) < self.target_selector:
            raise NotEnoughTargets()
        return self._inf_select(targets, self.target_selector)

    def _range_selector(self, targets: List):
        low_bound, high_bound = self.target_selector
        target_amount = len(targets)
        if target_amount < low_bound:
            raise NotEnoughTargets()
        return targets[:high_bound]

    def _function_selector(self, targets: List):
        return self.target_selector(targets)

    def _select(self, targets):
        if isinstance(self.target_selector, int) or math.isinf(self.target_selector):
            return self._limit_selector(targets)
        elif isinstance(self.target_selector, tuple):
            return self._range_selector(targets)
        elif isinstance(self.target_selector, Callable):
            return self._function_selector(targets)
        else:
            raise TypeError("Invalid target selector, must be an limit (int), range (tuple of 2 ints), "
                            "or a function (returns the targets or raises FailedDetection/NotEnoughTargets)")

    def select_targets(self, targets):
        return self._select(targets)

    def direct(self, targets, image: np.ndarray) -> Any:
        """
        Returns the directions using the `vision.director` for the given targets using the image.

        `Director.direct` works in the following way:
            - if the amount of targets is less than target amount then set directions to failed_detection
            - else it passes only target_selector of targets to the direction function
            - The final result (direction function or failed_detection) is passed to the direction monitors
            - each modify_directions is applied to the result and its result is passed to the next one
            - then the final result is returned

        :param targets: the list of object that were detected in your image
        :param image: the image from which the targets were taken
        :return: Depends on the directing function and the direction monitors,
                 usually a number or tuple
        """
        try:
            targets = self.select_targets(targets)
            directions = self.directions(targets, image)
            return self.apply_direction_monitors(directions, targets, image)
        except NotEnoughTargets:
            self.logger.debug("Not enough targets to direct")
            return self.failed_detection

    def apply_direction_monitors(self, directions: Any, targets: List[np.ndarray], image: np.ndarray) -> Any:
        """
        Applies the list of direction monitors one after the other

        Direction monitors are functions that receive directions targets image and camera settings
        and change it some way, like applying a PID feedback loop, returning a "stop" value based on some condition
        (like getting close enough to the target)

        :param directions: the raw directions returned from the directions function
        :param targets: the list of targets that passed all filters
        :param image: the image from which the targets were found
        :return: the final direction result
        """
        if self.direction_monitors:
            for direction_monitor in self.direction_monitors:
                directions = direction_monitor.modify_directions(directions, targets, image)
                if direction_monitor.priority:
                    return directions
        return directions



from functools import partial

import cv2

from ..shape_filter_constants import POLYGON_FILTER_PRECISION, POLYGON_FILTER_DEFAULT_SIDE_AMOUNT
from ..target_filter import target_filter
from ...ovl_math import geometry
from ...ovl_math.contours import contour_lengths_and_angles
from ...utils.types import RangedNumber


def validate_ratio_against_average(value, value_average, divisor, threshold):
    return 1 - (abs(value - value_average) / divisor) < threshold


def polygon_filter_average(items, precision: int = POLYGON_FILTER_PRECISION):
    return round(sum(items) / float(len(items)), precision)


@target_filter
def polygon_filter(contour_list, side_amount=POLYGON_FILTER_DEFAULT_SIDE_AMOUNT,
                   angle_deviation: RangedNumber(0, 1) = 0,
                   side_length_deviation: RangedNumber(0, 1) = 0,
                   polygon_fill_ratio: RangedNumber(0, 1) = 0.7,
                   polygon_angle_ratio: RangedNumber(0, 1) = 0,
                   approximation_coefficient: RangedNumber(0, 1) = 0.02):
    """
     A filter that Detects regular polygon of variable amount of sides, default is 6 (Hexagon).
     The filter uses 4 shape characteristics to perform the filtering,
     1. Validate that all sides do not deviate too much from the polygon_filter_average side in the contour
         angle
     2. Validate that all angles do not deviate too much from the polygon_filter_average angle
     3. Validate that angle polygon_filter_average is close to the angle of the polygon with the same amount of sides
     4. Validate that the length of all sides are close to the length of the side of a polygon of a similar size

    :param polygon_angle_ratio:
    :param contour_list: the list of contours to be filtered
    :param side_amount: the amount of sides the wanted polygon has. The Default is to detect a Hexagon
    :param angle_deviation: the minimum ratio between each angle and the polygon_filter_average angle and
    the polygon_filter_average angle and the
    target angle of a shape of side_amount
    :param polygon_fill_ratio: The minimum ratio between the contour's area and the target area
    :param side_length_deviation: The minimum ratio between the length of each side and the polygon_filter_average
    length and between the polygon_filter_average length.
    :param approximation_coefficient: the coefficient for the function cv2.approxPolyDP.
    :return: the list of contours (numpy ndarrays) that passed the filter and are regular polygon
    """
    if side_amount < 3:
        raise ValueError("A polygon must have at least 3 sides")
    output = []
    output_append = output.append
    for current_contour in contour_list:
        vertices, lengths, angles = contour_lengths_and_angles(current_contour, approximation_coefficient)
        if len(vertices) == side_amount:
            average_length = polygon_filter_average(lengths)
            length_validator = partial(validate_ratio_against_average,
                                       average=average_length,
                                       divisor=average_length,
                                       threshold=side_length_deviation)
            if all(map(length_validator, lengths)):
                average_angle = polygon_filter_average(angles)
                angle_validator = partial(validate_ratio_against_average,
                                          average=average_angle,
                                          divisor=average_angle,
                                          threshold=angle_deviation)
                if all(map(angle_validator, angles)):
                    target_angle = geometry.regular_polygon_angle(side_amount)
                    if not validate_ratio_against_average(average_angle,
                                                          target_angle,
                                                          target_angle,
                                                          polygon_angle_ratio):
                        fill_ratio = cv2.contourArea(current_contour) / geometry.polygon_area(average_length,
                                                                                              side_amount)
                        if polygon_fill_ratio <= fill_ratio <= 1 / polygon_fill_ratio:
                            output_append(current_contour)
    return output



import asyncio
import functools
import logging
from typing import Union, List, Any, Generator, Dict, Tuple

from numpy import ndarray

from .ambient_vision import AmbientVision
from ..utils.get_function_name import get_function_name
from ..utils.constants import BASE_LOGGER, MULTIVISION_LOGGER
from ..utils.types import VisionLike

logger = logging.getLogger(f"{BASE_LOGGER}.{MULTIVISION_LOGGER}")
NO_VISIONS_GIVEN_MESSAGE = "No visions were given to the MultiVision object."


class MultiVision:
    """
    An object used to switch between multiple vision objects.
    It does this by having a list of visions and a connection and network_location to update from
    So it can be automatically updated to swap to the desired Vision object
    use the start method to start an infinite loop that returns images and detections

    .. code-block:: python

        connection = ovl.NetworkTablesConnection(1937)

        vision1 = ovl.Vision(....)
        vision2 = ovl.Vision(....)
        vision3 = ovl.Vision(....)

        controller_network_connection = ovl.NetworkLocation(table_key="current_vision")

        multi_vision = ovl.MultiVision([vision1, vision2, vision3], connection)

        for directions, targets, image  in multi_vision.start():

            # do something with the generated data

            # like sending the data or displaying the targets

            multi_vision.send(directions)

            ovl.display_contours(image, targets, delay=1)

    """

    def __init__(self, visions: Union[List[VisionLike], Dict[Any, VisionLike]], default_vision_index=None):
        self.pre_iteration_func = None
        self.post_iteration_func = None

        self.visions = visions
        if len(self.visions) == 0:
            logger.error(NO_VISIONS_GIVEN_MESSAGE)
            raise ValueError(NO_VISIONS_GIVEN_MESSAGE)

        if isinstance(visions, list):
            default_vision_index = default_vision_index or 0
            self._validate_vision_index_list(default_vision_index)
        elif isinstance(visions, dict):
            default_vision_index = default_vision_index or list(visions.keys())[0]
            self._validate_vision_index_dictionary(default_vision_index)
        self.default_vision = self.visions[default_vision_index]
        self.current_vision = self.default_vision
        self.default_vision_index = default_vision_index
        self.upcoming_vision = None
        self._update_vision_func = None
        self._update_task = None

    @property
    @functools.lru_cache()
    def vision_amount(self) -> int:
        """
        Returns the amount of visions in the controller

        :return: the amount of visions
        """
        return len(self.visions)

    @property
    def is_ambient(self) -> bool:
        """
         Returns True if the current_vision vision is an Ambient vision object

        :return: if the current vision is an AmbientVision
        """
        return isinstance(self.current_vision, AmbientVision)

    def before_iteration(self, func):
        """
        Decorator that runs the function before each iteration of the MultiVision loop

        :param func: the function to run before each iteration
        """
        self.pre_iteration_func = func
        return func

    def after_iteration(self, func):
        """
        Decorator that runs the function before each iteration of the MultiVision loop

        :param func: the function to run before each iteration
        """
        self.post_iteration_func = func
        return func

    def vision_updater(self, update_vision):
        """
        This function is a decorator that can be used to set custom logic to when vision updating occurs, it will be
        called constantly, but will actually update only once a detection iteration finishes.
        """

        @functools.wraps(update_vision)
        async def _update_vision():
            logger.debug("Starting vision update task, task_id: %s", get_function_name(_update_vision))
            is_update_async = asyncio.iscoroutinefunction(update_vision)
            while True:
                current_vision = self.current_vision
                if is_update_async:
                    next_vision = await update_vision(current_vision)
                else:
                    next_vision = update_vision(current_vision)
                self.upcoming_vision = next_vision

        self._update_vision_func = _update_vision()
        return update_vision

    def stop_update_task(self):
        """
        Stops the update task if it is running
        """
        if self._update_task is not None:
            self._update_task.cancel()
        else:
            raise ValueError("No update task is running")

    def set_new_vision(self):
        self.current_vision = self.upcoming_vision or self.current_vision

    async def start(self) -> Generator[Tuple[Any, "ndarray", Any], None, None]:
        """
        Start an infinite generator that takes an image
        detects with the current vision and returns the list of targets the image and directions
        and should be used as follows:

        .. code-block:: python

            connection = ovl.NetworkTablesConnection(1937)

            vision1 = ovl.Vision(....)
            vision2 = ovl.Vision(....)
            vision3 = ovl.Vision(....)

            update_location = ovl.NetworkLocation(table_key="current_vision")

            multi_vision = ovl.MultiVision([vision1, vision2, vision3], connection, update_location)

            for directions, targets, image  in multi_vision.start():

                # do something with the generated data

                # like sending the data or displaying the targets

                multi_vision.send(directions)

                ovl.display_contours(image, targets, delay=1)

        Note: automatically updates AmbientVision's vision swapping (AmbientVision.update_vision())!

        :yields: targets, image and directions
        """
        self._update_task = asyncio.create_task(self._update_vision_func)
        while True:
            self.set_new_vision()

            if self.pre_iteration_func:
                self.pre_iteration_func()

            image = await asyncio.get_running_loop().run_in_executor(None, self.current_vision.get_image)
            targets, filtered_image = self.current_vision.detect(image)
            directions = self.current_vision.director.direct(targets, filtered_image)
            if self.is_ambient:
                self.current_vision.update_vision()
            yield directions, targets, filtered_image

            if self.post_iteration_func:
                self.post_iteration_func()

    def _validate_vision_index_list(self, index):
        """
        Validates the index is in the list of visions

        :param index: the index to validate
        :return: the index
        """
        return 0 <= index < len(self.visions)

    def _validate_vision_index_dictionary(self, index):
        """
        Validates the index is a key of a vision in the controller

        :param index: the index to validate
        :return: the index
        """
        return index in self.visions



import cv2

from ..predicate_target_filter import predicate_target_filter
from ...ovl_math.shape_fill_ratios import rectangle_fill_ratio_straight
from ...utils.types import RangedNumber


@predicate_target_filter
def vertical_rectangle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.7):
    """
    Receives a list of contours and returns only the ones that are approximately a vertical rectangle

    :param contour: contour to be evaluated
    :param min_area_ratio: minimum ratio between the area of the contours and the bounding shape
    :return: the contour list filtered.
    """
   
    fill_ratio, contour_width, contour_height = rectangle_fill_ratio_straight(contour)
    perimeter = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, 0.02 * perimeter, True)
    return fill_ratio > min_area_ratio and contour_width < contour_height and len(approximation) == 4



import cv2

from ..predicate_target_filter import predicate_target_filter
from ...ovl_math.shape_fill_ratios import circle_fill_ratio
from ...utils.types import RangedNumber


@predicate_target_filter
def constraining_circle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.80,
                               min_len_ratio: RangedNumber(0, 1) = 0.9):
    """
    Filters out contour which are not approximately circle, also limits by radius ratio (unlike circle_filter).

    :param contour: a contour (numpy.ndarray) to be filtered
    :param min_area_ratio: minimum ratio between the area of the enclosing
    circle and the contour (contour /enclosing circle)
    :param min_len_ratio: minimum ratio between the radius of the enclosing circle and the enclosing rectangle.
    :return: the list of contours that fit the criteria
    """
    fill_ratio, radius = circle_fill_ratio(contour)
    _, _, rectangle_width, rectangle_height = cv2.boundingRect(contour)
    radius_ratio = ((2 * radius) ** 2) / float(rectangle_width) * rectangle_height
    return min_len_ratio <= (radius_ratio ** 0.5) and min_area_ratio <= fill_ratio



import cv2

from ..predicate_target_filter import predicate_target_filter
from ...ovl_math.shape_fill_ratios import rotating_rectangle_fill_ratio
from ...utils.types import RangedNumber


@predicate_target_filter
def rotated_rectangle_filter(contour, min_area_ratio: RangedNumber(0, 1) = 0.8):
    """
    Receives a list of contours and returns only those that are approximately a rectangle regardless
    of the angle of rotation.

    :param contour: contour currently being evaluated
    :param min_area_ratio: the minimum ratio between the contour area and the bounding shape
    :return: the contour list filtered.
    """
    fill_ratio, _, _ = rotating_rectangle_fill_ratio(contour)
    contour_perimeter = cv2.arcLength(contour, True)
    approximation = cv2.approxPolyDP(contour, 0.02 * contour_perimeter, True)
    return fill_ratio > min_area_ratio and len(approximation) == 4



from setuptools import setup, find_packages

from ovl import __version__


with open("README.md", "r") as read_me:
    long_description = read_me.read()

PRECOMPILED_OPENCV = ["opencv-python"]
FRC_EXTRA_PACKAGES = ["pynetworktables"]

extra_requirements = {
    "cv": PRECOMPILED_OPENCV,
    "frc": FRC_EXTRA_PACKAGES,
}

setup(
    name="ovl",
    packages=find_packages(),
    version=__version__,
    license="apache-2.0",
    author="SerpentBit",
    author_email="ovl.contact.help@gmail.com",
    description="A modular and versatile Python package for computer vision"
                " object detection pipelines tailored for robotics applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SerpentBit/ovl",
    install_requires=["numpy"],
    extras_require=extra_requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows"
    ]
)



class InvalidCustomFilterError(Exception):
    pass


class MissingResourceError(Exception):
    pass


class VersionError(Exception):
    pass


class CameraError(Exception):
    pass


class ImageError(Exception):
    pass


class FailedDetection(Exception):
    pass


class NotEnoughTargets(Exception):
    pass



from ..partials.keyword_partial import keyword_partial

IMAGE_FILTERS = set()


def image_filter(image_filter_function):
    """
    A decorator used to pass parameters in two stages,
    run in a python console `help(ovl.target_filter)`
    or look at ovl.target_filter documentation for more information
    image_filter acts like target_filter except it receives an image (np.ndarray)
    applies changes on it and then returns it altered.
    functions decorated by image_filter can be passed to Vision(image_filters_)

    Changes that are applied can vary from  blurs, image rotations, translations, contrast and brightness alterations
    and color spaces conversions (HSV, RGB Greyscale conversions)

    Filters that change binary images (images that have been threshold by
    a threshold object such as ovl.Color ovl.MultiColor or ovl.Canny) like morphological
    functions (ex. erosion, dilation) count as image filters as well and should be
    decorated using image_filter.

    image filter example:

    .. code-block:: python

        @image_filter
        def rotate_by_angle(image, angle):
            (height, width) = image.shape[:2]
            center_xy = (width / 2, height / 2)
            rotation_matrix = cv2.getRotationMatrix2D(center_xy, -angle, 1.0)
            cos = np.abs(rotation_matrix[0, 0])
            sin = np.abs(rotation_matrix[0, 1])
            rotated_image_dimensions = (int((height * sin) + (width * cos)), int((height * cos) + (width * sin)))
            rotation_matrix[0, 2] += (rotated_image_dimensions[0] / 2) - center_xy[0]
            rotation_matrix[1, 2] += (rotated_image_dimensions[1] / 2) - center_xy[1]
            return cv2.warpAffine(image, rotation_matrix, rotated_image_dimensions)

    The filter can then be passed to a Vision object to be applied to the images
    in the pipeline.


    .. code-block:: python

        pipeline = Vision(image_filters=[rotate_by_angle(angle=180)]
    """
    filter_partial = keyword_partial(image_filter_function)
    IMAGE_FILTERS.add(filter_partial)
    return filter_partial



import dataclasses
from typing import Dict, Any

from .camera_calibration import CameraCalibration
from .camera_properties import CameraProperties


@dataclasses.dataclass
class CameraConfiguration:
    camera_properties: Dict[CameraProperties, Any]
    camera_calibration: CameraCalibration = None



from typing import NewType, Callable, Union

import math

TargetSelector = NewType("TargetSelector", Union[Callable, int, tuple])


def validate_target_selector(target_selector: TargetSelector):
    if not isinstance(target_selector, (Callable, int, tuple)):
        raise TypeError("Target selector must be an int, a tuple of 2 ints or a function that receives the targets,"
                        f"target_selector: {target_selector}")
    if isinstance(target_selector, int):
        return math.inf if target_selector <= 0 else target_selector
    if isinstance(target_selector, tuple):
        try:
            low, high = target_selector
            if not (0 <= low < high):
                raise ArithmeticError("If target selector is a tuple, low must be bigger or equal to 0,"
                                      f" and high must be bigger that low, target_selector:  {target_selector}")
            return target_selector
        except (TypeError, ValueError):
            raise TypeError(f"If target selector is a tuple it must be of length 2! target_selector: {target_selector}")
    if isinstance(target_selector, Callable):
        return target_selector
    return target_selector



from typing import Union


class DirectionModifier:
    @property
    def priority(self) -> Union[bool, float]:
        """
        This is a value that determines the Modifier's priority.
        Determines whether it skips following DirectionModifiers.

        """
        return False

    def modify_directions(self, directions, targets, image):
        """
        Use this function to define behaviour that is applied to the directions you calculate.
        This can range from:
            - deciding when to stop
            - correcting camera location in comparison to the robot turn axis (i.e. Indentation)
            - applying some sort of pid on directions
            - convert the values to different scale or unit
            - saving data for analysis
            - interpolation the data from the directions calculated
            - anything else you wish to apply as a "logical" change on object detection behaviour
        :param image: the image the pipeline ran on.
        :param targets: the list of targets found at the end of the filtration
        :param directions: the result of the direction function
        :return: the modified direction
        """
        raise NotImplementedError("When inheriting DirectionMonitor you must implement it!")
